import struct
from typing import Literal


def floor_decimal(value: float, digit: int):
    """소수점 아래 자리수 자르기"""
    if digit == 0:
        return int(value)
    else:
        power_of_ten = 10**digit
        result: float = int(value * power_of_ten) / power_of_ten
        return result


def floor_precision(value: float, digit: int):
    """
    유효숫자 개수 자르기

    Examples:
        >>> floor_presicion(value=-123.45, digit=4)
        -123.4

        >>> floor_presicion(value=-123.45, digit=7)
        -123.45

        >>> floor_presicion(value=-123.45, digit=2)
        -120

        >>> floor_presicion(value=-123, digit=2)
        -120
    """
    value_str = str(value)
    number_str = value_str.replace(".", "").replace("-", "")
    if len(number_str) <= digit:
        return value
    else:
        if "." not in value_str:
            integer_part = value_str
        else:
            integer_part, fractional_part = value_str.split(".")

        integer_part_number_str = integer_part.replace("-", "")
        integer_part_count = len(integer_part_number_str)

        if integer_part_count > digit:  # 123.45, 2
            result_str = integer_part_number_str[:digit] + "0" * (
                integer_part_count - digit
            )
        elif integer_part_count == digit:  # 123.45, 3
            result_str = integer_part_number_str
        else:  # 123.45, 4
            result_str = (
                integer_part_number_str
                + "."
                + fractional_part[: digit - integer_part_count]
            )

        if value < 0:
            result_str = "-" + result_str
        return float(result_str)


def int_to_bytes(byteorder: Literal["little", "big"], value: int, length: int = None):
    if not length:
        length = (value.bit_length() + 7) // 8
    bytes_value = value.to_bytes(
        length=length,
        byteorder=byteorder,
        signed=value < 0,
    )
    return bytes_value


def int_to_2bytes(value: int):
    bytes_value = [(value >> 8) & 0xFF, value & 0xFF]
    return bytes_value


def bytes_to_int(byteorder: Literal["little", "big"], value: bytes):
    int_value = int.from_bytes(value, byteorder=byteorder, signed=True)
    return int_value


def bytes_to_bool(byteorder: Literal["little", "big"], value: bytes):
    int_value = bytes_to_int(byteorder=byteorder, value=value)
    bool_value = int_value == 1
    return bool_value


def int_to_bits(bit_count: int, value: int):
    bits_dict = {bit: value & (1 << bit) != 0 for bit in range(bit_count)}
    return bits_dict


def bytes_to_bits(byteorder: Literal["little", "big"], bit_count: int, value: bytes):
    int_value = bytes_to_int(byteorder=byteorder, value=value)
    bits_dict = int_to_bits(bit_count=bit_count, value=int_value)
    return bits_dict


def bytes_to_float(value: bytes):
    float_value = struct.unpack("f", value)[0]
    return float_value


def float_to_bytes(value: float):
    bytes_value = struct.pack("f", value)
    return bytes_value


def bytes_to_double(value: bytes):
    double_value: float = struct.unpack("!d", value)[0]
    return double_value


def str_to_bytes(value: str):
    bytes_value = value.encode()
    return bytes_value


def bytes_to_str(value: bytes):
    str_value: str = str(value, "ascii")
    return str_value


def str_to_ascii(value: str):
    bytes_value = [ord(ch) for ch in value]
    return bytes_value


def bytes_to_hex(byteorder: Literal["little", "big"], value: bytes):
    result = [int_to_hex(byteorder=byteorder, value=v) for v in value]
    return result


def int_to_hex(byteorder: Literal["little", "big"], value: int, pad_length: int = 1):
    min_length = (value.bit_length() + 7) // 8
    length = max(min_length, pad_length)
    result = [
        format(v, f"02x")
        for v in value.to_bytes(length, byteorder=byteorder, signed=value < 0)
    ]
    if len(result) == 1:
        result = result[0]
    return result


if __name__ == "__main__":
    assert (result := floor_precision(value=-123.45, digit=4)) == -123.4, result
    assert (result := floor_precision(value=-123.45, digit=7)) == -123.45, result
    assert (result := floor_precision(value=-123.45, digit=2)) == -120, result
    assert (result := floor_precision(value=-123, digit=2)) == -120, result
    assert (result := floor_precision(value=123, digit=2)) == 120, result
    assert (result := floor_precision(value=100.0, digit=3)) == 100, result



from dataclasses import dataclass

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import Color, LocationType, LocationReference
from ablelabs.neon.utils.format_conversion import floor_precision


@dataclass
class Speed:
    unit: str
    unit_s: float = None
    rate: float = None

    def __str__(self) -> str:
        result = []
        if self.unit_s != None:
            result.append(f"{self.unit_s}{self.unit}/s")
        if self.rate != None:
            result.append(f"{floor_precision(self.rate * 100, digit=3)}%")
        return f"{' '.join(result)}"

    def __repr__(self) -> str:
        return self.__str__()

    # None인 속성 제외
    # def __repr__(self):
    #     field_strings = []
    #     for field in fields(self):
    #         value = getattr(self, field.name)
    #         if value is not None:
    #             field_strings.append(f"{field.name}={repr(value)}")
    #     field_string = ", ".join(field_strings)
    #     return f"{self.__class__.__name__}({field_string})"

    @staticmethod
    def from_mm(mm: float):
        return Speed(unit="mm", unit_s=mm)

    @staticmethod
    def from_rate(rate: float = 1.0):
        return Speed(unit="mm", rate=rate)


class FlowRate(Speed):
    @staticmethod
    def from_ul(ul: float):
        return Speed(unit="ul", unit_s=ul)

    @staticmethod
    def from_rate(rate: float = 1.0):
        return Speed(unit="ul", rate=rate)


@dataclass
class LedBarParam:
    color: Color = Color.NONE
    on_brightness_percent: int = None
    off_brightness_percent: int = None
    bar_percent: int = None
    blink_time_ms: int = None


@dataclass
class Location:
    location_type: LocationType
    location_number: int
    well: str
    reference: LocationReference
    offset: tuple[float] = (0, 0, 0)

    # def to(
    #     self,
    #     location_number: int = None,
    #     well: str = None,
    #     reference: LocationReference = None,
    #     offset: tuple[float] = None,
    # ):
    #     return Location(
    #         location_type=self.location_type,
    #         location_number=(
    #             location_number if location_number else self.location_number
    #         ),
    #         well=well if well else self.well,
    #         reference=reference if reference else self.reference,
    #         offset=offset if offset else self.offset,
    #     )

    def __str__(self) -> str:
        # inspect만으로는 원하는 형태로 만들기가 어려울 듯.
        # result = " ".join(
        #     [
        #         f"{name}={value}"
        #         for name, value in inspect.getmembers(self)
        #         if "__" not in name and not inspect.isfunction(value)
        #     ]
        # )
        # return result
        result = f"{self.location_type}"
        if self.location_number:
            result += f".{self.location_number}"
        if self.well:
            result += f" well={self.well}"
        if self.reference:
            result += f" reference={self.reference}"
        if self.offset and self.offset != (0, 0, 0):
            result += f" offset={self.offset}"
        return result

    def __repr__(self) -> str:
        return self.__str__()


def location(
    location_type: LocationType = LocationType.DECK,
    location_number: int = None,
    well: str = None,
    reference: LocationReference = None,
    offset: tuple[float] = (0, 0, 0),
):
    return Location(
        location_type=location_type,
        location_number=location_number,
        well=well,
        reference=reference,
        offset=offset,
    )


if __name__ == "__main__":
    flow_rate = FlowRate.from_rate()
    print(flow_rate)
    flow_rate = FlowRate.from_rate(0.12345)
    print(flow_rate)
    flow_rate = FlowRate.from_rate(0.012345)
    print(flow_rate)

    led_bar_param = LedBarParam()
    print(led_bar_param)

    loc = Location(
        location_type=LocationType.DECK,
        location_number=12,
        well="a6",
        reference=LocationReference.BOTTOM,
        offset=(1, 2, 3),
    )
    print(loc)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter


class ShakerAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.deck_module_shaker_initialize)
    async def initialize(self):
        pass

    @run_server_func(RobotRouter.deck_module_shaker_move_to_ready)
    async def move_to_ready(self):
        pass

    @run_server_func(RobotRouter.deck_module_shaker_shake)
    async def shake(self, rpm: int, acceleration_sec: int):
        pass

    @run_server_func(RobotRouter.deck_module_shaker_shake_off)
    async def shake_off(self):
        pass



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.enums import NUMBERING_ORDER


class LocationConversion:
    @staticmethod
    def get_row_column(
        row_count: int,
        col_count: int,
        numbering_order: NUMBERING_ORDER,
        number: int,
        base: int = 1,
    ):
        if numbering_order == NUMBERING_ORDER.LTR_TTB:
            row, column = divmod(number - base, col_count)
        elif numbering_order == NUMBERING_ORDER.LTR_BTT:
            row, column = divmod(number - base, col_count)
        elif numbering_order == NUMBERING_ORDER.TTB_LTR:
            # row, column = divmod(number - base, row_count)
            # row = row_count - 1 - row
            raise NotImplementedError()
        row += base
        column += base
        return row, column

    @staticmethod
    def get_well(row: int, column: int, base: int = 1):
        well_row = chr(ord("A") + row - base)
        well_col = column + 1 - base
        return f"{well_row}{well_col}"

    @staticmethod
    def _get_cell(
        row_count: int,
        col_count: int,
        numbering_order: NUMBERING_ORDER,
        row: int,
        col: int,
        base: int = 1,
    ):
        if numbering_order == NUMBERING_ORDER.LTR_TTB:
            cell = (row - base) * col_count + col
        elif numbering_order == NUMBERING_ORDER.LTR_BTT:
            cell = (row_count - 1 - row + base) * col_count + col
        elif numbering_order == NUMBERING_ORDER.TTB_LTR:
            # cell = (row - base) * col_count + col
            raise NotImplementedError()
        return cell

    @staticmethod
    def well_row_column(
        labware_row: int,
        labware_col: int,
        numbering_order: NUMBERING_ORDER,
        well_str: str,
        base: int = 1,
    ):
        well_index: int = LocationConversion.well_str_to_well_index(
            labware_row=labware_row,
            labware_col=labware_col,
            numbering_order=numbering_order,
            well_str=well_str,
            base=base,
        )
        return LocationConversion.get_row_column(
            row_count=labware_row,
            col_count=labware_col,
            numbering_order=numbering_order,
            number=well_index,
            base=base,
        )

    @staticmethod
    def well_str_to_well_index(
        labware_row: int,
        labware_col: int,
        numbering_order: NUMBERING_ORDER,
        well_str: str,
        base: int = 1,
    ):
        well_row = ord(well_str[0].upper()) - ord("A") + base
        well_col = int(well_str[1:]) - 1 + base
        return LocationConversion._get_cell(
            row_count=labware_row,
            col_count=labware_col,
            numbering_order=numbering_order,
            row=well_row,
            col=well_col,
            base=base,
        )

    @staticmethod
    def well_index_to_well_str(
        labware_row: int,
        labware_col: int,
        numbering_order: NUMBERING_ORDER,
        well_index: int,
        base: int = 1,
    ):
        row, column = LocationConversion.get_row_column(
            row_count=labware_row,
            col_count=labware_col,
            numbering_order=numbering_order,
            number=well_index,
            base=base,
        )
        return LocationConversion.get_well(row=row, column=column, base=base)

    # @staticmethod
    # def str_bool_list_to_index_list(
    #     row_count: int,
    #     col_count: int,
    #     value: list[str],
    #     numbering_order: NUMBERING_ORDER,
    #     base: int = 1,
    #     sort: bool = True,
    # ):
    #     row_columns: list[tuple[int, int]] = []
    #     for row, rows in enumerate(value):
    #         for column, cell in enumerate(rows):
    #             if cell == "1":
    #                 row_columns.append((row + base, column + base))
    #     result = [
    #         LocationConversion.get_cell(
    #             row_count=row_count,
    #             col_count=col_count,
    #             numbering_order=numbering_order,
    #             row=row,
    #             col=column,
    #             base=base,
    #         )
    #         for row, column in row_columns
    #     ]
    #     if sort:
    #         result.sort()
    #     return result


if __name__ == "__main__":

    class RegionSample:
        def __init__(
            self, row: int, col: int, numbering_order: NUMBERING_ORDER
        ) -> None:
            self.row = row
            self.col = col
            self.numbering_order = numbering_order

    deck = RegionSample(row=4, col=3, numbering_order=NUMBERING_ORDER.LTR_BTT)
    well_plate_96 = RegionSample(row=8, col=12, numbering_order=NUMBERING_ORDER.LTR_TTB)

    row_column = LocationConversion.get_row_column(
        row_count=deck.row,
        col_count=deck.col,
        numbering_order=deck.numbering_order,
        number=1,
    )
    print(f"deck(1) = {row_column}")

    row_column = LocationConversion.get_row_column(
        row_count=deck.row,
        col_count=deck.col,
        numbering_order=deck.numbering_order,
        number=4,
    )
    print(f"deck(4) = {row_column}")

    row_column = LocationConversion.get_row_column(
        row_count=deck.row,
        col_count=deck.col,
        numbering_order=deck.numbering_order,
        number=12,
    )
    print(f"deck(12) = {row_column}")

    row_column = LocationConversion.well_row_column(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_str="a1",
    )
    print(f"well_plate_96(a1) = {row_column}")

    row_column = LocationConversion.well_row_column(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_str="a12",
    )
    print(f"well_plate_96(a12) = {row_column}")

    row_column = LocationConversion.well_row_column(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_str="h12",
    )
    print(f"well_plate_96(h12) = {row_column}")

    row_column = LocationConversion.well_row_column(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_str="b3",
    )
    print(f"well_plate_96(b3) = {row_column}")

    well = LocationConversion.get_well(row=1, column=1)
    print(f"well(1,1) = {well}")

    well = LocationConversion.get_well(row=1, column=12)
    print(f"well(1,12) = {well}")

    well = LocationConversion.get_well(row=8, column=12)
    print(f"well(8,12) = {well}")

    well = LocationConversion.get_well(row=16, column=24)
    print(f"well(16,24) = {well}")

    well = LocationConversion.well_index_to_well_str(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_index=1,
    )
    print(f"well_plate_96(1) = {well}")

    well = LocationConversion.well_index_to_well_str(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_index=12,
    )
    print(f"well_plate_96(12) = {well}")

    well = LocationConversion.well_index_to_well_str(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_index=96,
    )
    print(f"well_plate_96(96) = {well}")

    well = LocationConversion.well_index_to_well_str(
        labware_row=well_plate_96.row,
        labware_col=well_plate_96.col,
        numbering_order=well_plate_96.numbering_order,
        well_index=15,
    )
    print(f"well_plate_96(15) = {well}")



import numpy as np


class LinearMap:
    @staticmethod
    def get_transform_matrix(
        src_points: list[list[float]],
        dst_points: list[list[float]],
    ):
        src = np.array(src_points)
        dst = np.array(dst_points)

        # 4x4 transformation matrix를 구하기 위해 homogeneous 좌표계를 이용
        src_hom = np.hstack((src, np.ones((4, 1))))
        dst_hom = np.hstack((dst, np.ones((4, 1))))

        # Least-squares solution을 이용하여 transformation matrix 계산
        transform_matrix, _ = np.linalg.lstsq(src_hom, dst_hom, rcond=None)[:2]
        return transform_matrix

    @staticmethod
    def transform(
        transform_matrix,
        src_points: list[float],
    ) -> list[float]:
        src = np.array(src_points)

        # homogeneous 좌표계로 변환한 뒤, transformation matrix를 곱함
        new_hom = np.hstack((src, 1))
        transformed_hom = np.dot(new_hom, transform_matrix)

        # homogeneous 좌표계에서 일반 좌표계로 변환
        dst = transformed_hom[:3] / transformed_hom[3]
        dst_points = list(dst)
        return dst_points



def internal_division(
    x1: float,
    y1: float,
    z1: float,
    x2: float,
    y2: float,
    z2: float,
    m: float,
    n: float,
):
    """
    두 점을 내분하는 점의 좌표를 계산하는 함수

    Parameters:
        m (float): 첫 번째 점에서의 비율
        n (float): 두 번째 점에서의 비율

    Returns:
        tuple: 내분점의 (x, y) 좌표
    """
    x = (m * x2 + n * x1) / (m + n)
    y = (m * y2 + n * y1) / (m + n)
    z = (m * z2 + n * z1) / (m + n)
    return x, y, z



from typing import Callable, Coroutine

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.common.alma.enums import ZGripper
from ablelabs.neon.common.alma.structs import Location


class TimeGripperAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.time_upper_module_gripper_move_block_to_inspector)
    async def move_block_to_inspector(self, push: bool) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_move_block_to_holder)
    async def move_block_to_holder(self) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_open_lid)
    async def open_lid(self, chip_holder_number: int, lid_holder_number: int) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_close_lid)
    async def close_lid(self, lid_holder_number: int, chip_holder_number: int) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_move_labware)
    async def move_labware(
        self,
        from_location: Location,
        to_location: Location,
        push: bool = False,
    ) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_move_lid)
    async def move_lid(self, from_location: Location, to_location: Location) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_pick_labware)
    async def pick_labware(self, location: Location) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_place_labware)
    async def place_labware(self, location: Location, push: bool = False) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_pick_lid)
    async def pick_lid(self, location: Location) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_place_lid)
    async def place_lid(self, location: Location) -> float:
        pass

    @run_server_func(RobotRouter.time_upper_module_gripper_move_to)
    async def move_to(
        self,
        location: Location,
        z_gripper: ZGripper,
        # task_after_z_up: Callable[[], Coroutine] = None,
        slow_z_offset: float = 0,
    ) -> float:
        pass



import asyncio
from typing import Optional, TypeVar, Generic

T = TypeVar("T")


class LongPollingGetter(Generic[T]):
    def __init__(self, initial_value: Optional[T] = None):
        self._value: Optional[T] = initial_value
        self._event = asyncio.Event()
        if initial_value is not None:
            self._event.set()

    async def get(self) -> T:
        await self._event.wait()
        self._event.clear()
        return self._value
        # try:
        #     await self._event.wait()
        # # except asyncio.CancelledError as e:
        # #     pass
        # finally:
        #     self._event.clear()
        #     return self._value

    def set(self, value: T) -> None:
        self._value = value
        self._event.set()

    def __add__(self, other: T):
        if self._value is None:
            raise ValueError("Value is not set yet")
        if not isinstance(other, type(self._value)):
            raise TypeError(f"Cannot add {type(other)} to {type(self._value)}")
        new_value = self._value + other
        self.set(new_value)
        return self


async def main():
    a: LongPollingGetter[int] = LongPollingGetter()
    b: LongPollingGetter[float] = LongPollingGetter()
    c: LongPollingGetter[str] = LongPollingGetter()

    async def get_a():
        while True:
            print(f"a = {await a.get()}")

    async def get_b():
        while True:
            print(f"b = {await b.get()}")

    async def get_c():
        while True:
            print(f"c = {await c.get()}")

    tasks = [
        asyncio.create_task(task)
        for task in [
            get_a(),
            get_b(),
            get_c(),
        ]
    ]

    a.set(1)
    await asyncio.sleep(1)
    b.set(2.3)
    await asyncio.sleep(1)
    c.set("c")
    await asyncio.sleep(1)
    a += 1
    b.set(2.33)
    c.set("cc")
    await asyncio.sleep(1)
    a.set(1)
    b.set(2.333)
    c.set("ccc")


if __name__ == "__main__":
    asyncio.run(main())



import cv2

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.kribb.api.robot_router import RobotRouter
from ablelabs.neon.common.notable.kribb.structs import LabelInfo, ColonyInfo


class OpticAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.optic_set_camera_live)
    async def set_camera_live(self, on: bool):
        pass

    @run_server_func(RobotRouter.optic_camera_capture)
    async def camera_capture(self, resize_ratio: float = None) -> cv2.Mat:
        pass

    @run_server_func(RobotRouter.optic_camera_show)
    async def camera_show(
        self,
        winname: str,
        resize_ratio: float = 0.5,
    ):
        pass

    @run_server_func(RobotRouter.optic_set_led_brightness)
    async def set_led_brightness(self, value: int):
        pass

    @run_server_func(RobotRouter.optic_set_led_on_off)
    async def set_led_on_off(self, on: bool):
        pass

    @run_server_func(RobotRouter.optic_set_displacement_zero)
    async def set_displacement_zero(self):
        pass

    @run_server_func(RobotRouter.optic_get_displacement_value)
    async def get_displacement_value(self) -> float:
        pass

    @run_server_func(RobotRouter.optic_detect_colony)
    async def detect_colony(
        self,
        image: cv2.Mat = None,
        resize_ratio: float = 0.5,
        padding: tuple[float, float, float, float] = (0.5, 0.5, 0.5, 0.5),
        rect_refer: cv2.typing.Rect = (8, 8, 83, 83),
        rect_tolerance: int = 5,
        vignette_size: int = 30,
        binary_thresh: int = 170,
        crop_border_thresh: int = 30,
        return_process_images: bool = False,
        font_scale: float = 0.7,
    ) -> tuple[tuple[int, int, int, int], list[LabelInfo], cv2.Mat, dict[str, cv2.Mat]]:
        pass

    @run_server_func(RobotRouter.optic_group_colony)
    async def group_colony(
        self,
        crop_info: tuple[int, int, int, int],
        label_infos: list[LabelInfo],
        image: cv2.Mat = None,
        pick_count: int = 1,
        resize_ratio: float = 0.5,
        padding: tuple[float, float, float, float] = (0.5, 0.5, 0.5, 0.5),
        initial_dist: float = 60,
        row: int = 8,
        column: int = 12,
    ) -> tuple[dict[int, list[LabelInfo]], cv2.Mat]:
        pass



from loguru import logger

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.nanophilia.api.get_api import GetAPI
from ablelabs.neon.controllers.notable.nanophilia.api.set_api import SetAPI

# from ablelabs.neon.controllers.notable.nanophilia.api.time_api import TimeAPI
from ablelabs.neon.controllers.notable.nanophilia.api.motion_api import MotionAPI
from ablelabs.neon.controllers.notable.nanophilia.api.deck_module import DeckModuleAPI
from ablelabs.neon.controllers.notable.nanophilia.api.axis_api import AxisAPI
from ablelabs.neon.common.notable.nanophilia.enums import RunStatus, DioInput


class RobotAPI(MessengerClient):
    def __init__(self) -> None:
        tcp_client = TcpClient(name="tcp_client", log_func=logger.trace)
        super().__init__(tcp_client)
        self._get_api = GetAPI(tcp_client=tcp_client)
        self._set_api = SetAPI(tcp_client=tcp_client)
        # self._time_api = TimeAPI(tcp_client=tcp_client)
        self._motion_api = MotionAPI(tcp_client=tcp_client)
        self._deck_module_api = DeckModuleAPI(tcp_client=tcp_client)
        self._axis_api = AxisAPI(tcp_client=tcp_client)

    @property
    def get(self):
        return self._get_api

    @property
    def set(self):
        return self._set_api

    # @property
    # def time(self):
    #     return self._time_api

    @property
    def motion(self):
        return self._motion_api

    @property
    def deck_module(self):
        return self._deck_module_api

    @property
    def axis(self):
        return self._axis_api

    async def connect(self, ip, port):
        await self._tcp_client.connect(ip=ip, port=port)

    @run_server_func(RobotRouter.robot_wait_boot)
    async def wait_boot(self):
        pass

    @run_server_func(RobotRouter.robot_shutdown, wait_return=False)
    async def shutdown(self):
        pass

    @run_server_func(RobotRouter.robot_stop)
    async def stop(self):
        pass

    @run_server_func(RobotRouter.robot_clear_error)
    async def clear_error(self):
        pass

    @run_server_func(RobotRouter.robot_pause)
    async def pause(self):
        pass

    @run_server_func(RobotRouter.robot_resume)
    async def resume(self):
        pass

    @run_server_func(RobotRouter.robot_is_connected)
    async def is_connected(self) -> dict[str, bool]:
        pass

    @run_server_func(RobotRouter.robot_get_dio_input)
    async def get_dio_input(self) -> dict[DioInput, bool]:
        pass

    @run_server_func(RobotRouter.robot_get_environment)
    async def get_environment(self) -> dict[str, float]:
        pass

    @run_server_func(RobotRouter.robot_get_run_status)
    async def get_run_status(self) -> RunStatus:
        pass

    @run_server_func(RobotRouter.robot_get_changed_robot_status)
    async def get_changed_robot_status(self) -> dict:
        pass

    @run_server_func(RobotRouter.robot_set_interlock)
    async def set_interlock(self, value: bool):
        pass

    @run_server_func(RobotRouter.robot_set_led_lamp)
    async def set_led_lamp(self, on: bool):
        pass

    @run_server_func(RobotRouter.robot_set_run_status)
    async def set_run_status(self, value: RunStatus):
        pass

    @run_server_func(RobotRouter.robot_set_progress_rate)
    async def set_progress_rate(self, value: float):
        pass

    @run_server_func(RobotRouter.robot_initialize)
    async def initialize(self):
        pass

    @run_server_func(RobotRouter.robot_delay)
    async def delay(self, sec: float):
        pass

    @run_server_func(RobotRouter.robot_get_ffu)
    async def get_ffu(self) -> dict:
        pass

    @run_server_func(RobotRouter.robot_set_ffu)
    async def set_ffu(self, on: bool, rpm: float = None):
        pass

    @run_server_func(RobotRouter.robot_set_ffu_alarm_clear)
    async def set_ffu_alarm_clear(self):
        pass



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter


class GetAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)

    @run_server_func(RobotRouter.get_setup_data)
    async def setup_data(self) -> dict:
        pass



# python 3.10.0

# Install the ablelabs package using "pip install ablelabs"
# Import necessary modules from ablelabs package
from ablelabs.neon.controllers.notable.api.robot_api import RobotAPI
from ablelabs.neon.common.notable.enums import (
    Axis,
    LocationReference,
    RunStatus,
    NUMBERING_ORDER,
)
from ablelabs.neon.common.notable.structs import Speed, FlowRate, location
from ablelabs.neon.utils.location_conversion import LocationConversion
from ablelabs.neon.utils.decorators import log_func_args_async


async def main():
    import subprocess

    # Launch Neon.exe using subprocess (for Windows systems)
    # Ensure 'Neon.exe' and the 'resources' folder are in the correct directory
    subprocess.Popen(
        [
            r"C:\your_path\Neon.exe",
        ]
    )

    # Instantiate the robot API client
    robot_api = RobotAPI()
    # Ensure to follow the sequence: connect -> wait_boot -> set_pipettes/tips/labwares -> initialize -> motion_api for proper operation

    # Connect to the robot system
    ip = "localhost"
    port = 1234
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass  # Handle exceptions for connection failures

    # Wait for the robot system to boot
    await robot_api.wait_boot()

    # --- Set up robot components ---

    # Configure pipettes
    await robot_api.set.pipettes(
        {
            1: "8ch200ul",
        }
    )

    # Configure tips
    await robot_api.set.tips(
        {
            1: "tip_200",
        }
    )

    # Configure labware layout on the deck
    await robot_api.set.labwares(
        {
            1: "12_reservoir",
            2: "96_rv_plate",
            3: "96_pcr_plate",
            10: "tipbox_200",
            11: "tipbox_200",
            12: "trash",
        }
    )

    # --- User Defined ---

    pipette_number = 1
    reservoir = 1
    rv_plate = 2
    pcr_plate = 3
    tipboxs = [10, 11]
    trash = 12
    tipbox_wells = [(tipbox, f"a{col+1}") for tipbox in tipboxs for col in range(12)]
    pipette_volume = 200
    wells = [f"a{col+1}" for col in range(12)]

    @log_func_args_async(log_before_func=print)
    async def pick_up_tip():
        tipbox, well = tipbox_wells.pop(0)
        print(f"{tipbox=} {well=}")
        await robot_api.motion.pick_up_tip(
            pipette_number=pipette_number,
            location=location(location_number=tipbox, well=well),
        )

    @log_func_args_async(log_before_func=print)
    async def drop_tip():
        await robot_api.motion.drop_tip(
            pipette_number=pipette_number,
            location=location(location_number=trash),
        )

    @log_func_args_async(log_before_func=print)
    async def aspirate(deck: int, well: str, volume: float):
        await robot_api.motion.aspirate(
            pipette_number=pipette_number,
            volume=volume,
            location=location(
                location_number=deck,
                well=well,
                reference=LocationReference.BOTTOM,
            ),
        )

    @log_func_args_async(log_before_func=print)
    async def dispense(deck: int, well: str, volume: float):
        await robot_api.motion.dispense(
            pipette_number=pipette_number,
            volume=volume,
            location=location(
                location_number=deck,
                well=well,
                reference=LocationReference.BOTTOM,
            ),
        )

    # --- Motion Commands ---

    # Initialize the motion system
    await robot_api.motion.initialize()

    # reservoir -> 96_rv_plate: 10ul, multi, reuse tip
    volume = 10
    chunk_count = int(pipette_volume / volume)
    chunked_wells = [
        wells[i : i + chunk_count] for i in range(0, len(wells), chunk_count)
    ]

    await pick_up_tip()
    for wells in chunked_wells:
        await aspirate(deck=reservoir, well="a1", volume=volume * len(wells))
        for well in wells:
            await dispense(deck=rv_plate, well=well, volume=volume)
    await drop_tip()

    # 96_rv_plate -> 96_pcr_plate: 10ul, single, change tip
    volume = 10
    for well in wells:
        await pick_up_tip()
        await aspirate(deck=rv_plate, well=well, volume=volume)
        await dispense(deck=pcr_plate, well=well, volume=volume)
        await drop_tip()


if __name__ == "__main__":
    # Run the main async function with an event loop
    import asyncio

    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



from dataclasses import dataclass
import cv2

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.common.notable.structs import *


@dataclass
class LabelInfo:
    index: int
    width: int
    height: int
    area: int
    centroid: tuple[float, float]
    inscribed_center: tuple[float, float]
    inscribed_radius: int
    perimeter: float
    circularity: float
    brightness: float
    color: tuple[float, float, float]
    contours: list[cv2.Mat]


@dataclass
class ColonyInfo:
    x_ratio: float
    y_ratio: float
    area: float
    circularity: float
    brightness: float
    color: tuple[float, float, float]



# python 3.10 버전

# 가상환경 설정
## python -m venv .venv
## py 3.10 -m venv .venv

# ablelabs 패키지 설치
## pip install ablelabs

# Neon.exe + resources 파일 위치 확인
## resources 폴더는 Neon.exe와 같은 폴더 또는 하위 폴더에

# subprocess.Popen으로 Neon.exe 실행

# connect -> wait_boot -> set_pipettes/tips/labwares 순서대로 호출


import asyncio
import sys, os
import cv2

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.controllers.notable.kribb.api.robot_api import RobotAPI
from ablelabs.neon.common.notable.enums import LocationType, Axis, LocationReference
from ablelabs.neon.common.notable.structs import Speed, FlowRate, location
from ablelabs.neon.utils.format_conversion import floor_decimal
from ablelabs.neon.utils.location_conversion import LocationConversion, NUMBERING_ORDER


async def main():
    import subprocess

    # for windows.
    # subprocess.Popen(
    #     [
    #         r"C:\Users\ABLE Labs\Desktop\ABLE-Elba\robot\dist\Neon.exe",
    #     ]
    # )

    # for mac.
    # subprocess.Popen(
    #     [
    #         r"/Users/sypark/Code/ABLE-Elba/.venv/bin/python",
    #         r"/Users/sypark/Code/ABLE-Elba/robot/src/controllers/notable/kribb/robot_router.py",
    #     ],
    #     cwd=r"/Users/sypark/Code/ABLE-Elba",
    # )
    # await asyncio.sleep(2)

    ip = "localhost"
    port = 1234

    robot_api = RobotAPI()
    try:
        await robot_api.connect(ip=ip, port=port)
    except Exception as e:
        pass

    await robot_api.wait_boot()

    # set
    await robot_api.set.pipettes(
        {
            # 1: "8ch1000ul",
            2: "1ch200ul",
        }
    )
    await robot_api.set.tips(
        {
            # 1: "tip_1000",
            2: "tip_200",
        }
    )
    await robot_api.set.camera(pipette_number=1)
    await robot_api.set.deck_modules(
        {
            10: "light_module",
        }
    )
    await robot_api.set.labwares(
        {
            12: "trash_#v2.5",
            11: "ablelabs_tiprack_#AL-CT-200",
            10: "spl_trayplate_60ml_#30001",
            7: "spl_96_well_0.2ml_#30096",
        }
    )

    # get
    pipette_infos = await robot_api.get.pipette_infos()
    tip_infos = await robot_api.get.tip_infos()
    deck_module_infos = await robot_api.get.deck_module_infos()
    labware_infos = await robot_api.get.labware_infos()
    setup_data = await robot_api.get.setup_data()

    x, y, z = setup_data["teaching"]["camera_1"]["deck10_up"]
    setup_data["teaching"]["camera_1"]["deck10_up"] = [x + 1, y + 2, z + 3]
    await robot_api.set.update_setup_data(setup_data, True)

    # robot
    await robot_api.stop()
    await robot_api.clear_error()  # stop 이후, motion 전에.
    await robot_api.pause()
    await robot_api.resume()
    is_connected = await robot_api.is_connected()

    # after motion.initialize
    await robot_api.motion.initialize()
    await robot_api.motion.move_to_camera(
        pipette_number=1,
        location=location(
            location_number=10,
            reference=LocationReference.BOTTOM_JUST,  # 동일하게 설정 필요.
        ),
    )
    (x_values, y_values, z_grid, width, height, scan_image) = (
        await robot_api.scan_displacement_sensor(
            pipette_number=1,
            location=location(
                location_number=10,
            ),
            padding=(5, 5),
            scan_count=(8, 5),
        )
    )
    cv2.imshow("scan_image", scan_image)
    # cv2.waitKey(0)
    await robot_api.move_to_interpolated_z(
        pipette_number=2,
        location=location(
            location_number=10,
            offset=(0, 0, 0),
            reference=LocationReference.BOTTOM,
        ),
        x=10,
        y=-15,  # 음수로 설정.
    )

    # motion
    await robot_api.motion.initialize()
    await robot_api.motion.move_to(
        pipette_number=2,
        location=location(
            location_number=7,
            well="a1",
            reference=LocationReference.BOTTOM,
            offset=(1.0, 2.0, 3.0),
        ),
    )

    await robot_api.motion.pick_up_tip(
        pipette_number=2,
        location=location(
            location_number=11,
            well="a1",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=2,
        location=location(
            location_number=11,
            well="a1",
        ),
    )
    await robot_api.motion.pick_up_tip(
        pipette_number=2,
        location=location(
            location_number=11,
            well="a12",
        ),
    )
    await robot_api.motion.drop_tip(
        pipette_number=2,
        location=location(
            location_number=12,
        ),
    )

    await robot_api.motion.aspirate(
        pipette_number=2,
        volume=20,
        location=location(
            location_number=7,
            well="a1",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.rise_tip(
        pipette_number=2,
        height_offset=5,
        z_speed=Speed.from_mm(2),
    )
    await robot_api.motion.dispense(
        pipette_number=2,
        volume=20,
        location=location(
            location_number=7,
            well="a2",
            reference=LocationReference.BOTTOM,
        ),
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.mix(
        pipette_number=2,
        volume=20,
        iteration=2,
        # location=location(
        #     location_number=7,
        #     well="a3",
        #     reference=LocationReference.BOTTOM,
        # ),
        flow_rate=FlowRate.from_ul(20),
        delay=0.1,
    )
    await robot_api.motion.blow_out(
        pipette_number=2,
        flow_rate=FlowRate.from_ul(20),
    )
    await robot_api.motion.move_to_ready()

    # axis
    position = await robot_api.axis.get_position(axis=Axis.X)  # mm
    await robot_api.axis.set_speed(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.set_accel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.set_decel(axis=Axis.X, value=10)  # mm/sec2
    await robot_api.axis.disable(axis=Axis.X)
    await robot_api.axis.enable(axis=Axis.X)
    await robot_api.axis.stop(axis=Axis.X)
    await robot_api.axis.home(axis=Axis.X)
    await robot_api.axis.wait_home_done(axis=Axis.X)
    await robot_api.axis.jog(axis=Axis.X, value=10)  # mm/sec
    await robot_api.axis.jog(axis=Axis.X, value=0)  # mm/sec
    await robot_api.axis.step(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)
    await robot_api.axis.move(axis=Axis.X, value=10)  # mm
    await robot_api.axis.wait_move_done(axis=Axis.X)

    # optic
    await robot_api.optic.set_camera_live(on=True)
    image: cv2.Mat = await robot_api.optic.camera_capture(resize_ratio=0.5)
    # await robot_api.optic.camera_show(winname="camera", resize_ratio=0.5)
    await robot_api.optic.set_camera_live(on=False)
    await robot_api.optic.set_led_brightness(value=3)
    await robot_api.optic.set_led_on_off(on=True)
    image: cv2.Mat = await robot_api.optic.camera_capture(resize_ratio=0.3)
    await robot_api.optic.set_led_on_off(on=False)
    await robot_api.optic.set_displacement_zero()
    z = await robot_api.optic.get_displacement_value()
    image = cv2.imread(
        "/Users/sypark/Code/ABLE-Elba/robot/tests/notable_kribb/images/240904.bmp"
    )
    # cv2.imshow("image", image)
    (crop_info, label_infos, detected_image, process_images) = await robot_api.optic.detect_colony(
        image=image,  # or None (to use last captured image)
        resize_ratio=0.5,
        padding=(0.5, 0.5, 0.5, 0.5),
        rect_refer=(8, 8, 83, 83),
        rect_tolerance=5,
        vignette_size=30,
        binary_thresh=170,
        crop_border_thresh=30,
        return_process_images=True,
        font_scale=0.7,  # or None (to do not add index)
    )
    if process_images:
        cv2.imshow("vignette_image", process_images["vignette_image"])
        cv2.imshow("normalized_image", process_images["normalized_image"])
        cv2.imshow("binary_image", process_images["binary_image"])
        cv2.imshow("crop_border_image", process_images["crop_border_image"])
    for label_info in label_infos:
        info = [
            label_info.index,
            label_info.width,
            label_info.height,
            label_info.area,
            label_info.centroid,
            label_info.inscribed_center,
            label_info.inscribed_radius,
            label_info.perimeter,
            label_info.circularity,
            label_info.brightness,
            label_info.color,
            label_info.contours,
        ]
    cv2.imshow("detected_image", detected_image)
    (picked_infos, groupped_infos, groupped_image) = await robot_api.optic.group_colony(
        crop_info=crop_info,
        label_infos=label_infos,
        image=image,  # or None (to use last captured image)
        pick_count=3,
        resize_ratio=0.5,
        padding=(0.5, 0.5, 0.5, 0.5),
        initial_dist=60,
        row=8,
        column=12,
    )
    cv2.imshow("groupped_image", groupped_image)
    crop_x, crop_y, crop_w, crop_h = crop_info
    agar_plate_info: dict = labware_infos[10]
    well_offset_x, well_offset_y = (9, 6)
    xys = []
    for group_number, picked_info in dict(picked_infos).items():
        if not picked_info:
            continue
        for info in picked_info:
            x_pixel, y_pixel = info.inscribed_center
            x = (x_pixel / crop_w * agar_plate_info["well_x"]) + well_offset_x
            y = (y_pixel / crop_h * agar_plate_info["well_y"]) + well_offset_y
            xys.append((x, -y))  # y는 음수로 설정.
    target_plate_info: dict = labware_infos[7]
    well_index = 1
    for x, y in xys:
        well_row, well_column = LocationConversion.get_row_column(
            row_count=target_plate_info["row_count"],
            col_count=target_plate_info["column_count"],
            numbering_order=NUMBERING_ORDER.LTR_TTB,
            number=well_index,
        )
        well = LocationConversion.get_well(row=well_row, column=well_column)
        await robot_api.motion.move_to(
            pipette_number=2,
            location=location(
                location_number=10,
                reference=LocationReference.BOTTOM_JUST,
                offset=(x, y, -1.6),
            ),
            optimize=False,
        )
        await robot_api.motion.mix(
            pipette_number=2,
            volume=2,
            iteration=2,
            location=location(
                location_number=7,
                well=well,
                reference=LocationReference.BOTTOM,
                offset=(0, 0, 1),
            ),
        )
        well_index += 1

    cv2.waitKey(0)

    import io

    success, encoded_image = cv2.imencode(".png", image)
    # if not success:
    #     raise HTTPException(status_code=500, detail="Image encoding failed")
    image_bytes = io.BytesIO(encoded_image.tobytes())
    # return StreamingResponse(image_bytes, media_type="image/png")


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    loop.create_task(main())
    loop.run_forever()



import asyncio
from loguru import logger
import pickle
from typing import Callable, Any
import binascii
from functools import update_wrapper
import datetime
import traceback
import cv2
import numpy as np

import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.utils.network.tcp_server import TcpServer
from ablelabs.neon.utils.decorators import get_args_dict

IP = "localhost"
PORT = 2345


class Struct:
    def __init__(self, a: int, b: float, c: str) -> None:
        self.a = a
        self.b = b
        self.c = c

    def __str__(self) -> str:
        return f"a={self.a} b={self.b} c={self.c}"


class Messenger:
    ENCODING = "utf-8"

    @staticmethod
    def decode(packet):
        packet_bytes = pickle.dumps(packet)
        packet_hex = binascii.hexlify(packet_bytes)
        return packet_hex.decode(Messenger.ENCODING)
        # return packet_bytes.hex()

    @staticmethod
    def encode(msg: str):
        packet_bytes = binascii.unhexlify(msg)
        return pickle.loads(packet_bytes)

    @staticmethod
    def mat_to_png(image: cv2.Mat):
        retval, buf = cv2.imencode(".png", image)
        return buf

    @staticmethod
    def png_to_mat(buf: np.ndarray[np.uint8]):
        image = cv2.imdecode(buf, cv2.IMREAD_COLOR)
        return image

    @staticmethod
    def convert_args_format(args: dict) -> dict:
        """
        cv2.Mat -> PNG
        PNG -> cv2.Mat
        """
        converted_args = {}
        for key, value in args.items():
            if isinstance(value, cv2.Mat):  # cv2.Mat -> PNG
                converted_args[key] = Messenger.mat_to_png(value)
            elif (
                isinstance(value, np.ndarray)
                and value[:8].tobytes() == b"\x89PNG\r\n\x1a\n"  # png_signature
            ):  # PNG -> cv2.Mat
                converted_args[key] = Messenger.png_to_mat(value)
            else:
                converted_args[key] = value
        return converted_args

    @staticmethod
    def to_request_packet(time: datetime.datetime, func: Callable, args: dict):
        args = Messenger.convert_args_format(args)
        packet = (time, func.__name__, args)
        result = Messenger.decode(packet)
        return result

    @staticmethod
    def to_request_tuple(msg: str):
        packet = Messenger.encode(msg)
        (time, func_name, args) = packet
        args = Messenger.convert_args_format(args)
        return (time, func_name, args)

    @staticmethod
    def to_response_packet(time: datetime.datetime, func: Callable, args: dict, rtn):
        args = Messenger.convert_args_format(args)
        packet = (time, func.__name__, args, rtn)
        result = Messenger.decode(packet)
        return result

    @staticmethod
    def to_response_tuple(msg: str):
        packet = Messenger.encode(msg)
        (time, func_name, args, rtn) = packet
        args = Messenger.convert_args_format(args)
        return (time, func_name, args, rtn)


class MessengerServer:
    def __init__(self, tcp_server: TcpServer) -> None:
        self._tcp_server = tcp_server

        async def on_received(msg: str):
            asyncio.create_task(self.execute(msg))

        self._tcp_server.on_received.append(on_received)

    async def execute(self, msg: str):
        (time, func_name, args) = Messenger.to_request_tuple(msg)
        try:
            func = getattr(self, func_name)
        except AttributeError as e:
            send_msg = Messenger.to_response_packet(
                time, func_name, args, NotImplementedError()
            )
        else:
            try:
                rtn = await func(**args)
            except Exception as e:
                e_tb = MessengerServer.extract_traceback(e)
                send_msg = Messenger.to_response_packet(time, func, args, e_tb)
            else:
                send_msg = Messenger.to_response_packet(time, func, args, rtn)
        await self._tcp_server.send(send_msg)

    @staticmethod
    def extract_traceback(e: Exception):
        tb = traceback.extract_tb(e.__traceback__)
        frame = tb[-1]
        filename, lineno, function, text = frame
        filename = str(filename).split("\\")[-1]
        result = type(e)(f"{e}, {filename}, line {lineno}, in {function}, {text}")
        return result


class MessengerClient:
    def __init__(self, tcp_client: TcpClient) -> None:
        self._tcp_client = tcp_client
        self._requests = []

        async def on_received(msg: str):
            if len(self._requests) == 0:
                return
            (time, func_name, args, rtn) = Messenger.to_response_tuple(msg)
            for req in self._requests:
                (req_time, req_func_name, req_args, req_futre) = req
                if (
                    time == req_time
                    and func_name == req_func_name
                    and pickle.dumps(args) == pickle.dumps(req_args)
                ):
                    future: asyncio.Future = req_futre
                    future.set_result(rtn)
                    # try:
                    #     future.set_result(rtn)
                    # except asyncio.exceptions.InvalidStateError as e:
                    #     logger.warning(f"{e} : func={func_name} args={args} rtn={rtn}")
                    self._requests.remove(req)
                # else:
                #     print(f"{time=}")
                #     print(f"{req_time=}")
                #     print(f"{func_name=}")
                #     print(f"{req_func_name=}")
                #     print(f"{args=}")
                #     print(f"{req_args=}")

        self._tcp_client.on_received.append(on_received)

    async def send(self, time: datetime.datetime, func: Callable, args: dict):
        msg = Messenger.to_request_packet(time, func, args)
        await self._tcp_client.send(msg)

    async def request(self, func: Callable, args: dict, wait_return: bool):
        if not self._tcp_client.is_connected():
            raise ConnectionError(f"{MessengerClient.__name__} not connected.")
            # return
        req_time = datetime.datetime.now()
        future = asyncio.Future()
        self._requests.append((req_time, func.__name__, args, future))
        await self.send(time=req_time, func=func, args=args)
        if wait_return:
            await asyncio.gather(future)
            return future.result()
            # try:
            #     await asyncio.gather(future)
            # except InterruptedError as e:
            #     return False
            # # except asyncio.exceptions.InvalidStateError as e:
            # #     logger.error(f"{e}")
            # else:
            #     return future.result()


def run_server_func(
    origin_func: Callable = None,
    wait_return: bool = True,
):
    def decorator(func: Callable):
        async def wrapper(*args, **kwargs):
            _func = origin_func if origin_func else func
            if args and isinstance(args[0], MessengerClient):
                self = args[0]
                args = args[1:]
                args_dict = get_args_dict(_func, *args, **kwargs)
                result = await self.request(
                    func=_func,
                    args=args_dict,
                    wait_return=wait_return,
                )
                if isinstance(result, Exception):
                    e_args = {
                        "e": f"{result}",
                        "func": _func.__qualname__,
                        "args": args_dict,
                    }
                    e_args_str = f"{e_args}"
                    # e_args_str = get_dict_str(e_args)
                    e = type(result)(e_args_str)
                    logger.error(e)
                    # raise e from result
                    if isinstance(result, InterruptedError):
                        return e
                    else:
                        raise e from result
                else:
                    return result

        # update_wrapper(wrapper, func)
        return wrapper

    return decorator


class CustomMessengerServer(MessengerServer):
    async def func_a(self, args: Struct):
        import datetime

        logger.info(f"func_a({args}) - 1")
        await asyncio.sleep(1)
        logger.info(f"func_a({args}) - 2")
        return f"args={args} {datetime.datetime.now()}"

    async def func_a2(self, args: Struct):
        import datetime

        logger.info(f"func_b({args}) - 1")
        await asyncio.sleep(1)
        logger.info(f"func_b({args}) - 2")
        return f"args={args} {datetime.datetime.now()}"

    async def func_exception(self):
        a = 0 / 0


class CustomMessengerClient(MessengerClient):
    @run_server_func()
    async def func_a(self, args: Struct):
        pass

    @run_server_func(CustomMessengerServer.func_a2)
    async def func_b(self, args: Struct):
        pass

    @run_server_func(CustomMessengerServer.func_exception)
    async def func_exception(self):
        pass


async def main():
    logger.remove()
    # logger.add(sys.stdout, level="TRACE")
    logger.add(sys.stdout, level="DEBUG", backtrace=False)
    # logger.add(sys.stdout, level="INFO")
    logger.add("logs/trace.log", level="TRACE")
    logger.add("logs/debug.log", level="DEBUG")
    logger.add("logs/info.log", level="INFO")

    tcp_server = TcpServer()
    tcp_client = TcpClient()
    await tcp_server.open(ip=IP, port=PORT)
    await tcp_client.connect(ip=IP, port=PORT)
    server = CustomMessengerServer(tcp_server)
    client = CustomMessengerClient(tcp_client)
    # logger.info(await client.func_a(args=Struct(a=1, b=2.345, c="c")))
    # logger.info(await client.func_a(args=Struct(a=2, b=2.345, c="cc")))
    # logger.info(await client.func_a(args=Struct(a=3, b=2.345, c="ccc")))
    # logger.info(await client.func_a(args=Struct(a=4, b=2.345, c="cccc")))
    # logger.info(await client.func_b(args=Struct(a=1, b=2.345, c="c")))
    # logger.info(await client.func_b(args=Struct(a=2, b=2.345, c="cc")))
    # logger.info(await client.func_b(args=Struct(a=3, b=2.345, c="ccc")))
    # logger.info(await client.func_b(args=Struct(a=4, b=2.345, c="cccc")))

    # logger.info("start")
    # task = asyncio.create_task(client.func_a(args=Struct(a=9, b=9.9, c="9")))
    # task2 = asyncio.create_task(client.func_a(args=Struct(a=9, b=9.9, c="9")))
    # results = await asyncio.gather(task, task2)
    # logger.info(f"end : {results}")

    tasks = [
        asyncio.create_task(async_func)
        for async_func in [
            client.func_a(args=Struct(a=9, b=9.9, c="9")),
            client.func_b(args=Struct(a=99, b=99.9, c="99")),
            client.func_exception(),
        ]
    ]
    while not all([task.done() for task in tasks]):
        print("waiting...")
        await asyncio.sleep(0.1)
    results = [task.result() for task in tasks]
    logger.info(f"end : {results}")


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    # loop.set_debug(True)
    loop.create_task(main())
    loop.run_forever()



import asyncio
from typing import Callable
from loguru import logger
import json


ACK = "[ACK]"
ETX = "\x03"


class BaseTcp:
    def __init__(
        self,
        name: str,
        encoding: str = "utf-8",
        log_func: Callable = logger.trace,
    ) -> None:
        self._name = name
        self._encoding = encoding
        self._log_func = log_func
        self._is_connected = False
        self._reader: asyncio.StreamReader = None
        self._writer: asyncio.StreamWriter = None
        self.on_connected: list[Callable[[],]] = []
        self.on_received: list[Callable[[str],]] = []

    async def disconnect(self):
        if not self.is_connected():
            return
        self._is_connected = False
        if self._writer:
            try:
                self._writer.close()
                await self._writer.wait_closed()
            except Exception as e:
                logger.exception(f"{e}")
                pass

    def is_connected(self):
        # return self._sock is not None and self._is_connected
        return self._is_connected

    async def send(self, message, do_log: bool = True):
        if not self.is_connected():
            return
        if do_log and self._log_func:
            self._log_func(f"[{self._name}] -> {message}")
        message = self._check_json_format(message)
        byte_message = bytearray(f"{message}{ETX}", self._encoding)
        try:
            self._writer.write(byte_message)
            await self._writer.drain()
        except asyncio.CancelledError as e:
            logger.warning(f"Exception: {e}")
            await self.disconnect()

    @staticmethod
    def _check_json_format(value: str | dict):
        if type(value) == str:
            try:
                value_json = json.loads(value)
                result = json.dumps(value_json)
            except:
                result = value
        elif type(value) == dict:
            result = json.dumps(value)
        else:
            result = value
        return result

    # def request(self, message):
    #     self.

    async def handle_receive(self):
        input_buffer = ""
        while self.is_connected():
            try:
                input = await self._reader.read(1024)
                if not input:
                    await self.disconnect()
                input_buffer += input.decode(self._encoding)
                while ETX in input_buffer:
                    idx = input_buffer.find(ETX)
                    temp = input_buffer[:idx]
                    input_buffer = input_buffer[idx + 1 :]

                    message = temp.split(ETX)
                    for m in message:
                        if m.find(ACK) == 0:
                            continue
                        if self._log_func:
                            self._log_func(f"[{self._name}] <- {m}")
                        await self.send(f"{ACK}{m}", do_log=False)
                        for on_received in self.on_received:
                            await on_received(m)
            except (
                ConnectionResetError,
                ConnectionAbortedError,
                asyncio.IncompleteReadError,
            ) as e:
                logger.warning(f"Exception: {e}")
                await self.disconnect()
                return
            except Exception as e:
                logger.exception(f"Exception: {e}", e)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.alma.api.robot_router import RobotRouter
from ablelabs.neon.controllers.alma.api.deck_modules.inspector import InspectorAPI
from ablelabs.neon.controllers.alma.api.deck_modules.heater import HeaterAPI


class DeckModuleAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)
        self.inspector = InspectorAPI(tcp_client)
        self.heater = HeaterAPI(tcp_client)



import sys, os

sys.path.append(os.path.abspath(os.curdir))
from ablelabs.neon.utils.network.messenger import MessengerClient, run_server_func
from ablelabs.neon.utils.network.tcp_client import TcpClient
from ablelabs.neon.controllers.notable.nanophilia.api.robot_router import RobotRouter
from ablelabs.neon.controllers.notable.nanophilia.api.deck_modules.washer_dryer import (
    WasherDryerAPI,
)
from ablelabs.neon.controllers.notable.nanophilia.api.deck_modules.magnetic_shaker import (
    MagneticShakerAPI,
)
from ablelabs.neon.controllers.notable.nanophilia.api.deck_modules.shaker import (
    ShakerAPI,
)


class DeckModuleAPI(MessengerClient):
    def __init__(self, tcp_client: TcpClient) -> None:
        super().__init__(tcp_client)
        self.washer_dryer = WasherDryerAPI(tcp_client)
        self.magnetic_shaker = MagneticShakerAPI(tcp_client)
        self.shaker = ShakerAPI(tcp_client)
